package com.bergerkiller.bukkit.nolagg;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.minecraft.server.Entity;
import net.minecraft.server.EntityExperienceOrb;
import net.minecraft.server.EntityItem;
import net.minecraft.server.WorldServer;

import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;
import org.bukkit.util.Vector;

import com.bergerkiller.bukkit.common.utils.EntityUtil;

public class WorldStackFormer {

	private Object lock = new Object();
	private boolean orbrunning = false;
	private boolean itemrunning = false;
	private final List<EntityItem> items = new ArrayList<EntityItem>();
	private final List<EntityExperienceOrb> orbs = new ArrayList<EntityExperienceOrb>();
	private Set<EntityItem> itemsToRespawn = new HashSet<EntityItem>();

	public void update(List<Item> itemlist, List<ExperienceOrb> orblist) {
		if (!ItemHandler.formStacks) return;
		if (StackFormer.stackThreshold < 2) return;
		synchronized (lock) {
			if (itemrunning || orbrunning) return;
			for (EntityItem item : itemsToRespawn) {
				ItemHandler.respawnItem((Item) item.getBukkitEntity(), new Vector());
			}
			itemsToRespawn.clear();
			if (NoLagg.stackOrbs) {
				for (ExperienceOrb orb : orblist) {
					orbs.add((EntityExperienceOrb) EntityUtil.getNative(orb));
				}
				orbrunning = true;
			}

			for (Item item : itemlist) {
				items.add((EntityItem) EntityUtil.getNative(item));
			}
			itemrunning = true;
		}		
	}

	@SuppressWarnings("unchecked")
	private void updateOrbs() {
		for (EntityExperienceOrb orb : orbs) {
			if (orb.dead) continue;
			if (addSameOrbsNear(orbs, orb)) {
				if (near.size() > StackFormer.stackThreshold - 2) {
					for (EntityExperienceOrb to : (List<EntityExperienceOrb>) near) {
						if (to.dead) continue;
						//add the experience
						orb.value += to.value;
						kill(to);
					}						
				}
				near.clear();
			}
		}
		orbs.clear();
	}	

	private void kill(Entity entity) {
		((WorldServer) entity.world).tracker.untrackEntity(entity);
		entity.dead = true;
	}

	@SuppressWarnings("unchecked")
	private void updateItems() {
		int maxsize;
		for (EntityItem item : items) {
			if (item.dead) continue;
			maxsize = item.itemStack.getMaxStackSize();
			if (item.itemStack.count >= maxsize) continue;
			if (addSameItemsNear(items, item)) {
				if (near.size() > StackFormer.stackThreshold - 2) {
					//addition the items
					boolean changed = false;
					for (EntityItem nearitem : (List<EntityItem>) near) {
						if (nearitem.dead) continue;
						if (nearitem.itemStack == null) continue;
						if (nearitem.itemStack.count >= maxsize) continue;
						int newamount = item.itemStack.count + nearitem.itemStack.count;
						if (newamount <= maxsize) {
							if (newamount == 0) newamount = 1;
							item.itemStack.count = newamount;
							kill(nearitem);
							itemsToRespawn.remove(nearitem);
						} else {
							//set to max
							item.itemStack.count = maxsize;
							//set prev. item
							nearitem.itemStack.count = newamount - maxsize;
						}
						changed = true;
					}
					if (changed) {
						itemsToRespawn.add(item);
					}
				}
				near.clear();
			}
		}
		items.clear();
	}

	@SuppressWarnings("rawtypes")
	private static List near = new ArrayList(StackFormer.stackThreshold - 1);

	@SuppressWarnings("unchecked")
	private static boolean addSameItemsNear(List<EntityItem> from, EntityItem around) {
		if (around.dead) return false;
		for (EntityItem item : from) {
			if (item.dead) continue;
			if (item == around) continue;
			if (item.itemStack.id == around.itemStack.id && item.itemStack.b == around.itemStack.b) {
				if (canStack(around, item)) {
					near.add(item);
				}
			}
		}
		return !near.isEmpty();
	}

	@SuppressWarnings("unchecked")
	private static boolean addSameOrbsNear(List<EntityExperienceOrb> from, EntityExperienceOrb around) {
		if (around.dead) return false;
		for (EntityExperienceOrb orb : from) {
			if (orb.dead) continue;
			if (orb == around) continue;
			if (canStack(around, orb)) {
				near.add(orb);
			}
		}
		return !near.isEmpty();
	}
	private static boolean canStack(Entity e1, Entity e2) {
		double d = distance(e1.locX, e2.locX);
		if (d > StackFormer.stackRadiusSquared) return false;
		d += distance(e1.locZ, e2.locZ);
		if (d > StackFormer.stackRadiusSquared) return false;
		d += distance(e1.locY, e2.locY);
		if (d > StackFormer.stackRadiusSquared) return false;
		return true;
	}

	private static double distance(double d1, double d2) {
		d1 = Math.abs(d1 - d2);
		return d1 * d1;
	}

	public void run() {
		boolean running;
		synchronized (lock) {
			running = itemrunning;
		}
		if (running) {
			updateItems();
			synchronized (lock) {
				itemrunning = false;
			}
		}

		if (NoLagg.stackOrbs) {
			synchronized (lock) {
				running = orbrunning;
			}
			if (running) {
				updateOrbs();
				synchronized (lock) {
					orbrunning = false;
				}
			}
		}
	}

}
